#! /usr/bin/env python
# -*- coding: utf-8 -*-

'''demo.py: Contains a sample demo of the nearest wheelchair package.'''

import sys
import time

import nearest_wheelchair as nw


def main():
    #
    # Find nearest wheelchair using the non-Euclidean space dataset
    #
    # Create a graph
    floors = 2
    print 'Creating graph of AHL with %s floors...' % floors
    ahl_graph = nw.get_noneuclidean_dataset(floors=floors)
    ahl_graph.set_random_travel_times()  # Set edge weights

    # Place n=20 wheelchairs
    n = 20
    print 'Placing %s wheelchairs...' % n
    wheelchairs = ahl_graph.place_n_wheelchairs(n)

    # Create a porter -> patient job
    porter = ahl_graph.select_random_node()  # Source
    patient = ahl_graph.select_random_node()  # Target
    print 'Porter: [%s], Patient: [%s]' % (porter, patient)

    # Find the nearest wheelchair with respect to the path between the porter
    # and the patient using bi-directional Dijkstra's algorithm
    start_time = time.time()
    print 'Bi-directional Dijkstra Algorithm'
    qp = nw.Quickest_path(ahl_graph, wheelchairs)
    results = qp.find_quickest_path(qp.bidirectional_dijkstra_algorithm,
                                    porter, patient)
    print 'Runtime %s seconds' % ('{0:.3f}'.format(time.time()-start_time))
    print 'Explored nodes: %s' % results[0]['explored_nodes']
    print 'Path length and distance: %s, %s' % (len(results[0]['path']),
                                                '{0:.3f}'.format(results[0]['dist']))

    print
    print

    #
    # Find nearest wheelchair using the Euclidean space dataset
    #
    # Create a graph
    nodes = 400
    radius = 0.5
    print 'Creating a graph with %s nodes sampled from a circle with %s radius...'\
          % (nodes, radius)
    random_graph = nw.get_euclidean_dataset(n=nodes, radius=radius)

    # Build m landmarks
    m = 10
    print 'Building %s landmarks using Eigenvector centrality selection...' % m
    start_time = time.time()
    landmarks = nw.Landmarks(random_graph)
    landmarks.build_landmarks(m, nw.centrality_selection)
    print 'Building %s landmarks took %s seconds'\
          % (m, '{0:.3f}'.format(time.time()-start_time))

    # Place n=40 wheelchairs
    n = 40
    print 'Placing %s wheelchairs...' % n
    wheelchairs = random_graph.place_n_wheelchairs(n)

    # Create a porter -> patient job
    porter = random_graph.select_random_node()  # Source
    patient = random_graph.select_random_node()  # Target
    print 'Porter: [%s], Patient: [%s]' % (porter, patient)

    # Find the nearest wheelchair with respect to the path between the porter
    # and the patient using bi-directional A* algorithm
    heuristic = nw.landmark_heuristic_1
    start_time = time.time()
    print 'Bi-directional A* Algorithm using the triangle inequality heuristic'
    qp = nw.Quickest_path(random_graph, wheelchairs, landmarks)
    results = qp.find_quickest_path(qp.bidirectional_astar_algorithm,
                                    porter, patient, heuristic)
    print 'Runtime %s seconds' % ('{0:.3f}'.format(time.time()-start_time))
    print 'Explored nodes: %s' % results[0]['explored_nodes']
    print 'Path length and distance: %s, %s' % (len(results[0]['path']),
                                                '{0:.3f}'.format(results[0]['dist']))


if __name__ == '__main__':
    main()

    # Safe exit
    sys.exit(0)
