#! /usr/bin/env python
# -*- coding: utf-8 -*-

'''experiments.py: Contains functions for testing and plotting the nearest
wheelchair package.'''

from __future__ import division
import pickle
import sys
import time

import matplotlib.pyplot as plt
from mpltools import special
from mpltools import style
from mpltools import layout
import numpy as np

import nearest_wheelchair as nw

style.use('ggplot')

NUM_VERTS = 203  #203 is the base number of nodes in the AHL graph with 1 floor
GRAPH_SIZE = 0  # Increase graph size iteratively
WHEELCHAIRS = 1  # Increase the number of wheelchairs iteratively
EUCLIDEAN_DATASET = 2
NONEUCLIDEAN_DATASET = 3


def landmark_selection(graph, max_landmarks, max_searches):
    '''Measures and plots the search and landmark creation times.'''
    lur, sur = landmark_selection_test(graph, max_landmarks, max_searches,
                                       nw.uniform_selection)

    ldr, sdr = landmark_selection_test(graph, max_landmarks, max_searches,
                                       nw.degree_selection)

    lecr, secr = landmark_selection_test(graph, max_landmarks, max_searches,
                                         nw.centrality_selection)

    lcr, scr = landmark_selection_test(graph, max_landmarks, max_searches,
                                       nw.coverage_selection)

    # Search runtime plot
    sur_mean = [nw.statistics.pmean(sur[s]) for s in sur]
    sur_err = np.array([nw.statistics.variance(sur[s]) for s in sur])
    sdr_mean = [nw.statistics.pmean(sdr[s]) for s in sdr]
    sdr_err = np.array([nw.statistics.variance(sdr[s]) for s in sdr])
    secr_mean = [nw.statistics.pmean(secr[s]) for s in secr]
    secr_err = [nw.statistics.variance(secr[s]) for s in secr]
    scr_mean = [nw.statistics.pmean(scr[s]) for s in scr]
    scr_err = [nw.statistics.variance(scr[s]) for s in scr]

    figsize = layout.figaspect(scale=1.0)
    fig, axes = plt.subplots(ncols=1, nrows=1, figsize=figsize)
    axes.set_title('A* Runtime')
    axes.set_ylabel('Runtime (seconds)')
    axes.set_xlabel('Number of Landmarks')

    special.errorfill(range(0, max_landmarks), np.array(sur_mean), sur_err,
                      label='Uniform Selection', alpha_fill=0.1, ax=axes)
    special.errorfill(range(0, max_landmarks), np.array(sdr_mean), sdr_err,
                      label='Degree Selection', alpha_fill=0.1, ax=axes)
    special.errorfill(range(0, max_landmarks), np.array(secr_mean), secr_err,
                      label='Centrality Selection', alpha_fill=0.1, ax=axes)
    special.errorfill(range(0, max_landmarks), np.array(scr_mean), scr_err,
                      label='Coverage Selection', alpha_fill=0.1, ax=axes)

    layout.cross_spines(ax=axes)
    plt.legend()

    # Landmark runtime plot
    figsize1 = layout.figaspect(scale=1.0)
    fig1, axes1 = plt.subplots(ncols=1, nrows=1, figsize=figsize1)
    axes1.set_title('Landmark Creation Runtime')
    axes1.set_ylabel('Runtime (seconds)')
    axes1.set_xlabel('Number of Landmarks')

    ls = plt.rcParams['lines.linestyle']
    lw = plt.rcParams['lines.linewidth']
    color = axes1._get_lines.color_cycle.next()
    axes1.plot(range(0, max_landmarks), lur, color, linestyle=ls, linewidth=lw,
               label='Uniform Selection')
    color = axes1._get_lines.color_cycle.next()
    axes1.plot(range(0, max_landmarks), ldr, color, linestyle=ls, linewidth=lw,
               label='Degree Selection')
    color = axes1._get_lines.color_cycle.next()
    axes1.plot(range(0, max_landmarks), lecr, color, linestyle=ls, linewidth=lw,
               label='Centrality Selection')
    color = axes1._get_lines.color_cycle.next()
    axes1.plot(range(0, max_landmarks), lcr, color, linestyle=ls, linewidth=lw,
               label='Coverage Selection')

    layout.cross_spines(ax=axes1)
    plt.legend()

    # Landmark runtime plot
    figsize2 = layout.figaspect(scale=1.0)
    fig2, axes2 = plt.subplots(ncols=1, nrows=1, figsize=figsize2)
    axes2.set_title('Landmark Creation Runtime')
    axes2.set_ylabel('Runtime (seconds)')
    axes2.set_xlabel('Number of Landmarks')

    ls = plt.rcParams['lines.linestyle']
    lw = plt.rcParams['lines.linewidth']
    color = axes2._get_lines.color_cycle.next()
    axes2.plot(range(0, max_landmarks), lur, color, linestyle=ls, linewidth=lw,
               label='Uniform Selection')
    color = axes2._get_lines.color_cycle.next()
    axes2.plot(range(0, max_landmarks), ldr, color, linestyle=ls, linewidth=lw,
               label='Degree Selection')

    layout.cross_spines(ax=axes2)
    plt.legend()

    plt.show()


def landmark_selection_test(graph, max_landmarks, max_searches, algorithm):
    '''Measures both landmark selection time and the runtime of the regular ALT
    for an increasing amount of landmarks.'''
    landmark_runtimes = []
    search_runtimes = {}
    heuristic = nw.landmark_heuristic_1

    for i in range(1, max_landmarks+1):
        # Create landmarks
        print 'Building Landmarks (%s)...' % i
        start_time = time.time()
        landmarks = nw.Landmarks(graph)
        landmarks.build_landmarks(i, algorithm)
        landmark_runtimes.append(time.time()-start_time)

        # Perform search using the landmarks
        search_runtimes[i] = []
        for j in range(max_searches):
            source = graph.select_random_node()
            target = graph.select_random_node()
            start_time = time.time()
            nw.astar.astar(graph, source, target, heuristic, landmarks)
            search_runtimes[i].append(time.time()-start_time)
    print

    return landmark_runtimes, search_runtimes


def optimality(graph, landmarks, max_searches, num_wheelchairs):
    '''Measures and plots the optimality of all search algorithms.'''
    # Get and print error data
    results = optimality_test(graph, max_searches, num_wheelchairs, landmarks)

    # Plot errors in a bar chart
    x = np.arange(8)
    y1 = [results['DL'][0], results['ALT1'][0], results['ALT2'][0],
         results['AED'][0], results['BDD'][0], results['BDALT1'][0],
         results['BDALT2'][0], results['BDAED'][0]]
    y2 = [results['DL'][1], results['ALT1'][1], results['ALT2'][1],
         results['AED'][1], results['BDD'][1], results['BDALT1'][1],
         results['BDALT2'][1], results['BDAED'][1]]

    figsize = layout.figaspect(scale=1.0)
    fig, axes = plt.subplots(ncols=1, nrows=1, figsize=figsize)
    axes.set_title('Inaccuracies')
    axes.set_ylabel('Normalized Number of Errors')
    axes.set_xlabel('Algorithms')
    width = 1.0
    axes.bar(x, y1, width)
    axes.set_xticks(x+width*0.5)
    axes.set_xticklabels(['DL', 'A*TE1', 'A*TE2', 'A*ED', 'BDD', 'BDA*TE1',
                          'BDA*TE2', 'BDA*ED'])
    layout.cross_spines(ax=axes)

    # Plot distance errors in a bar chart
    figsize1 = layout.figaspect(scale=1.0)
    fig1, axes1 = plt.subplots(ncols=1, nrows=1, figsize=figsize1)
    axes1.set_title('Distance Errors')
    axes1.set_ylabel('Normalized Cumulative Distance Errors')
    axes1.set_xlabel('Algorithms')
    width = 1.0
    axes1.bar(x, y2, width, color=plt.rcParams['axes.color_cycle'][2])
    axes1.set_xticks(x+width*0.5)
    axes1.set_xticklabels(['DL', 'A*TE1', 'A*TE2', 'A*ED', 'BDD', 'BDA*TE1',
                          'BDA*TE2', 'BDA*ED'])

    layout.cross_spines(ax=axes1)


    plt.show()


def optimality_test(graph, max_searches, num_wheelchairs, landmarks):
    '''Tests every implemented algorithm with the quickest path algorithm and
    returns a count of their success/optimality. Comparisons are done with
    respect to Dijkstra's algorithm. The graph must be able to run every
    heuristic type (use random geometric graph)'''
    def check_optimality(optimal, solution):
        '''Quick check to see if the solution is optimal.'''
        errors = 0.0
        distance_error = 0.0
        for wh1, wh2 in zip(optimal, solution):
            if not (len(wh1['path']) == len(wh2['path'])) and (wh1['dist'] < wh2['dist']):
                errors += 1.0
                distance_error += abs(wh2['dist']-wh1['dist'])

        return errors/len(optimal), distance_error/len(optimal)

    optimality = {'DL': [0.0, 0.0], 'ALT1': [0.0, 0.0],
                  'ALT2': [0.0, 0.0], 'AED': [0.0, 0.0], 'BDD': [0.0, 0.0],
                  'BDALT1': [0.0, 0.0], 'BDALT2': [0.0, 0.0], 'BDAED': [0.0, 0.0]}

    # Run the simulation
    for i in range(max_searches):
        # Update wheelchair placements
        graph.reset_wheelchair_placement()
        wheelchairs = graph.place_n_wheelchairs(num_wheelchairs)
        qp = nw.Quickest_path(graph, wheelchairs, landmarks)

        # Get a source and target node and an optimal solution we can compare with
        source = graph.select_random_node()
        target = graph.select_random_node()
        optimal_solution = qp.find_quickest_path(qp.dijkstra_algorithm,
                                                 source, target)

        # Run every algorithm
        dl = qp.find_quickest_path(qp.dijkstra_landmarks_algorithm, source,
                                   target)
        alt1 = qp.find_quickest_path(qp.astar_algorithm, source, target,
                                     nw.landmark_heuristic_1)
        alt2 = qp.find_quickest_path(qp.astar_algorithm, source, target,
                                     nw.landmark_heuristic_2)
        aed = qp.find_quickest_path(qp.astar_algorithm, source, target,
                                    nw.euclidean_distance)
        bdd = qp.find_quickest_path(qp.bidirectional_dijkstra_algorithm,
                                    source, target)
        bdalt1 = qp.find_quickest_path(qp.bidirectional_astar_algorithm,
                                                source, target,
                                                nw.landmark_heuristic_1)
        bdalt2 = qp.find_quickest_path(qp.bidirectional_astar_algorithm,
                                                source, target,
                                                nw.landmark_heuristic_2)
        bdaed = qp.find_quickest_path(qp.bidirectional_astar_algorithm,
                                                source, target,
                                                nw.euclidean_distance)

        # Add to existing errors (errors/number of wheelchairs)
        res1, res2 = check_optimality(optimal_solution, dl)
        optimality['DL'][0] += res1; optimality['DL'][1] += res2
        res1, res2 = check_optimality(optimal_solution, alt1)
        optimality['ALT1'][0] += res1; optimality['ALT1'][1] += res2
        res1, res2 = check_optimality(optimal_solution, alt2)
        optimality['ALT2'][0] += res1; optimality['ALT2'][1] += res2
        res1, res2 = check_optimality(optimal_solution, aed)
        optimality['AED'][0] += res1; optimality['AED'][1] += res2
        res1, res2 = check_optimality(optimal_solution, bdd)
        optimality['BDD'][0] += res1; optimality['BDD'][1] += res2
        res1, res2 = check_optimality(optimal_solution, bdalt1)
        optimality['BDALT1'][0] += res1; optimality['BDALT1'][1] += res2
        res1, res2 = check_optimality(optimal_solution, bdalt2)
        optimality['BDALT2'][0] += res1; optimality['BDALT2'][1] += res2
        res1, res2 = check_optimality(optimal_solution, bdaed)
        optimality['BDAED'][0] += res1; optimality['BDAED'][1] += res2

    return optimality


def run_simulation(dataset_size, graph_type, n_landmarks, selection_algorithm,
                   iterator_type, max_searches, max_num_wheelchairs,
                   datasets=None, increment=None):
    '''Runs the simulation (see function below) and plots the results (runtime,
    explored vertices, and explored search space.'''
    # Create datasets
    if not datasets:
        print 'Creating datasets...'
        datasets = create_n_datasets(graph_type, dataset_size, n_landmarks,
                                     selection_algorithm)

    # Run simulation
    print 'Running simulation...'
    results = simulation(datasets, iterator_type, max_searches,
                         max_num_wheelchairs, increment)

    print 'Plotting the results...'
    # Plot the runtime results
    figsize = layout.figaspect(scale=1.0)
    fig, axes = plt.subplots(ncols=1, nrows=1, figsize=figsize)
    if NONEUCLIDEAN_DATASET == graph_type:
        axes.set_title('Runtime: Non-Euclidean Dataset')
    elif EUCLIDEAN_DATASET == graph_type:
        axes.set_title('Runtime: Euclidean Dataset')
    axes.set_ylabel('Runtime (seconds)')
    if GRAPH_SIZE == iterator_type:
        axes.set_xlabel('Dataset Size')
        x = range(1, dataset_size+1)
    elif WHEELCHAIRS == iterator_type:
        axes.set_xlabel('Number of Wheelchairs')
        #x = range(1, max_num_wheelchairs+1)
        x = range(0, max_num_wheelchairs+1, increment)
        x[0] = 1
        print x

    ls = plt.rcParams['lines.linestyle']
    lw = plt.rcParams['lines.linewidth']
    color = axes._get_lines.color_cycle.next()
    axes.plot(x, results['D']['runtimes'], color, linestyle=ls, linewidth=lw,
              label="Dijkstra")
    color = axes._get_lines.color_cycle.next()
    axes.plot(x, results['DL']['runtimes'], color, linestyle=ls, linewidth=lw,
              label="Dijkstra with Landmarks")
    color = axes._get_lines.color_cycle.next()
    axes.plot(x, results['ALT1']['runtimes'], color, linestyle=ls,
              linewidth=lw, label="A* TE1")
    color = axes._get_lines.color_cycle.next()
    axes.plot(x, results['ALT2']['runtimes'], color, linestyle=ls,
              linewidth=lw, label="A* TE2")
    if EUCLIDEAN_DATASET == graph_type:
        color = axes._get_lines.color_cycle.next()
        axes.plot(x, results['AED']['runtimes'], color, linestyle=ls,
                  linewidth=lw, label="A* ED")
    color = axes._get_lines.color_cycle.next()
    axes.plot(x, results['BDD']['runtimes'], color, linestyle=ls,
              linewidth=lw, label="BD Dijkstra")
    color = axes._get_lines.color_cycle.next()
    axes.plot(x, results['BDALT1']['runtimes'], color, linestyle=ls,
              linewidth=lw, label="BD A* TE1")
    color = axes._get_lines.color_cycle.next()
    axes.plot(x, results['BDALT2']['runtimes'], color, linestyle=ls,
              linewidth=lw, label="BD A* TE2")
    if EUCLIDEAN_DATASET == graph_type:
        color = axes._get_lines.color_cycle.next()
        axes.plot(x, results['BDAED']['runtimes'], color, linestyle=ls,
                  linewidth=lw, label="BD A* ED")

    layout.cross_spines(ax=axes)
    plt.xticks(x)
    #plt.xticks(range(1, int(plt.xticks()[0][-1])+2, increment))
    plt.legend(loc='upper left')

    # Plot the number of explored vertices
    figsize1 = layout.figaspect(scale=1.0)
    fig1, axes1 = plt.subplots(ncols=1, nrows=1, figsize=figsize1)
    if NONEUCLIDEAN_DATASET == graph_type:
        axes1.set_title('Explored Vertices: Non-Euclidean Dataset')
    elif EUCLIDEAN_DATASET == graph_type:
        axes1.set_title('Explored Vertices: Euclidean Dataset')
    axes1.set_ylabel('Number of Explored Vertices')
    if GRAPH_SIZE == iterator_type:
        axes1.set_xlabel('Dataset Size')
        x = range(1, dataset_size+1)
    elif WHEELCHAIRS == iterator_type:
        axes1.set_xlabel('Number of Wheelchairs')
        x = range(0, max_num_wheelchairs+1, increment)
        x[0] = 1

    ls = plt.rcParams['lines.linestyle']
    lw = plt.rcParams['lines.linewidth']
    color = axes1._get_lines.color_cycle.next()
    axes1.plot(x, results['D']['explored_nodes'], color, linestyle=ls,
               linewidth=lw, label="Dijkstra")
    color = axes1._get_lines.color_cycle.next()
    axes1.plot(x, results['DL']['explored_nodes'], color, linestyle=ls,
               linewidth=lw, label="Dijkstra with Landmarks")
    color = axes1._get_lines.color_cycle.next()
    axes1.plot(x, results['ALT1']['explored_nodes'], color, linestyle=ls,
               linewidth=lw, label="A* TE1")
    color = axes1._get_lines.color_cycle.next()
    axes1.plot(x, results['ALT2']['explored_nodes'], color, linestyle=ls,
               linewidth=lw, label="A* TE2")
    if EUCLIDEAN_DATASET == graph_type:
        color = axes1._get_lines.color_cycle.next()
        axes1.plot(x, results['AED']['explored_nodes'], color, linestyle=ls,
                   linewidth=lw, label="A* ED")
    color = axes1._get_lines.color_cycle.next()
    axes1.plot(x, results['BDD']['explored_nodes'], color, linestyle=ls,
               linewidth=lw, label="BD Dijkstra")
    color = axes1._get_lines.color_cycle.next()
    axes1.plot(x, results['BDALT1']['explored_nodes'], color, linestyle=ls,
               linewidth=lw, label="BD A* TE1")
    color = axes1._get_lines.color_cycle.next()
    axes1.plot(x, results['BDALT2']['explored_nodes'], color, linestyle=ls,
               linewidth=lw, label="BD A* TE2")
    if EUCLIDEAN_DATASET == graph_type:
        color = axes1._get_lines.color_cycle.next()
        axes1.plot(x, results['BDAED']['explored_nodes'], color, linestyle=ls,
                   linewidth=lw, label="BD A* ED")

    layout.cross_spines(ax=axes1)
    plt.xticks(x)
    plt.legend(loc='upper left')


    # Plot the explored search space
    figsize2 = layout.figaspect(scale=1.0)
    fig2, axes2 = plt.subplots(ncols=1, nrows=1, figsize=figsize2)
    if NONEUCLIDEAN_DATASET == graph_type:
        axes2.set_title('Explored Search Space: Non-Euclidean Dataset')
    elif EUCLIDEAN_DATASET == graph_type:
        axes2.set_title('Explored Search Space: Euclidean Dataset')
    axes2.set_ylabel('Explored Search Space')
    if GRAPH_SIZE == iterator_type:
        axes2.set_xlabel('Dataset Size')
        x = range(1, dataset_size+1)
    elif WHEELCHAIRS == iterator_type:
        axes2.set_xlabel('Number of Wheelchairs')
        x = range(0, max_num_wheelchairs+1, increment)
        x[0] = 1

    ls = plt.rcParams['lines.linestyle']
    lw = plt.rcParams['lines.linewidth']
    color = axes2._get_lines.color_cycle.next()
    axes2.plot(x, results['D']['explored_space'], color, linestyle=ls,
               linewidth=lw, label="Dijkstra")
    color = axes2._get_lines.color_cycle.next()
    axes2.plot(x, results['DL']['explored_space'], color, linestyle=ls,
               linewidth=lw, label="Dijkstra with Landmarks")
    color = axes2._get_lines.color_cycle.next()
    axes2.plot(x, results['ALT1']['explored_space'], color, linestyle=ls,
               linewidth=lw, label="A* TE1")
    color = axes2._get_lines.color_cycle.next()
    axes2.plot(x, results['ALT2']['explored_space'], color, linestyle=ls,
               linewidth=lw, label="A* TE2")
    if EUCLIDEAN_DATASET == graph_type:
        color = axes2._get_lines.color_cycle.next()
        axes2.plot(x, results['AED']['explored_space'], color, linestyle=ls,
                   linewidth=lw, label="A* ED")
    color = axes2._get_lines.color_cycle.next()
    axes2.plot(x, results['BDD']['explored_space'], color, linestyle=ls,
               linewidth=lw, label="BD Dijkstra")
    color = axes2._get_lines.color_cycle.next()
    axes2.plot(x, results['BDALT1']['explored_space'], color, linestyle=ls,
               linewidth=lw, label="BD A* TE1")
    color = axes2._get_lines.color_cycle.next()
    axes2.plot(x, results['BDALT2']['explored_space'], color, linestyle=ls,
               linewidth=lw, label="BD A* TE2")
    if EUCLIDEAN_DATASET == graph_type:
        color = axes2._get_lines.color_cycle.next()
        axes2.plot(x, results['BDAED']['explored_space'], color, linestyle=ls,
                   linewidth=lw, label="BD A* ED")

    layout.cross_spines(ax=axes2)
    plt.xticks(x)
    plt.legend(loc='upper left')

    plt.show()


def create_n_datasets(graph_type, n, m, landmark_algorithm):
    '''Creates n graphs of increasing size with m landmarks each.'''
    datasets = []
    if EUCLIDEAN_DATASET == graph_type:
        for i in range(1, n+1):
            print 'Dataset: %s' % i
            graph = nw.get_euclidean_dataset(n=NUM_VERTS*i, radius=0.05*i)
            landmarks = nw.Landmarks(graph)
            landmarks.build_landmarks(m, landmark_algorithm)
            datasets.append((graph, landmarks, EUCLIDEAN_DATASET))
    elif NONEUCLIDEAN_DATASET == graph_type:
        for i in range(1, n+1):
            print 'Dataset: %s' % i
            graph = nw.get_noneuclidean_dataset(floors=i)
            graph.set_random_travel_times()
            landmarks = nw.Landmarks(graph)
            landmarks.build_landmarks(m, landmark_algorithm)
            datasets.append((graph, landmarks, NONEUCLIDEAN_DATASET))

    return datasets


def simulation(datasets, iterator_type, max_searches, max_num_wheelchairs,
               increment=None):
    '''Runs a simulation of the nearest wheelchair package. Can be used
    to get runtime, explored vertices, and explored search space information.'''
    def create_data_structure(graph_type):
        dst = {'D': {}, 'DL': {}, 'ALT1': {}, 'ALT2': {}, 'BDD': {},
               'BDALT1': {}, 'BDALT2': {}}
        dst['D']['runtimes'] = []
        dst['D']['explored_nodes'] = []
        dst['D']['explored_space'] = []
        dst['DL']['runtimes'] = []
        dst['DL']['explored_nodes'] = []
        dst['DL']['explored_space'] = []
        dst['ALT1']['runtimes'] = []
        dst['ALT1']['explored_nodes'] = []
        dst['ALT1']['explored_space'] = []
        dst['ALT2']['runtimes'] = []
        dst['ALT2']['explored_nodes'] = []
        dst['ALT2']['explored_space'] = []
        dst['BDD']['runtimes'] = []
        dst['BDD']['explored_nodes'] = []
        dst['BDD']['explored_space'] = []
        dst['BDALT1']['runtimes'] = []
        dst['BDALT1']['explored_nodes'] = []
        dst['BDALT1']['explored_space'] = []
        dst['BDALT2']['runtimes'] = []
        dst['BDALT2']['explored_nodes'] = []
        dst['BDALT2']['explored_space'] = []
        if EUCLIDEAN_DATASET == graph_type:
            dst['AED'] = {'runtimes': [], 'explored_nodes': [],
                          'explored_space': []}
            dst['BDAED'] = {'runtimes': [], 'explored_nodes': [],
                            'explored_space': []}
        return dst

    # Can test runtime for an increasing graph size or increasing number of
    # wheelchairs
    num_wheelchairs = 0
    graph = None
    landmarks = None
    iterations = 0
    start = 1
    inc = 1
    if GRAPH_SIZE == iterator_type:
        iterations = len(datasets)
        num_wheelchairs = int(max_num_wheelchairs)
    elif WHEELCHAIRS == iterator_type:
        iterations = max_num_wheelchairs
        start = 0
        inc = int(increment)

    graph, landmarks, graph_type = datasets[-1]  # Use the largest graph

    # Return data structure: 'algorithm': {runtime, explored_nodes}
    results = create_data_structure(graph_type)

    # Run the simulation
    for i in range(start, iterations+1, inc):
        # Check if we should use a new graph or update the number of wheelchairs
        if GRAPH_SIZE == iterator_type:
            graph, landmarks, graph_type = datasets[i-1]
        elif WHEELCHAIRS == iterator_type:
            if not i:
                num_wheelchairs = 1
            else:
                num_wheelchairs = i

        print 'Simulation step: %s' % i

        # Run a user specified number of searches
        temp_results = create_data_structure(graph_type)
        for j in range(max_searches):
            # Update the wheelchair placements
            graph.reset_wheelchair_placement()
            wheelchairs = graph.place_n_wheelchairs(num_wheelchairs)
            qp = nw.Quickest_path(graph, wheelchairs, landmarks)

            # Get a source and target node
            source = graph.select_random_node()
            target = graph.select_random_node()

            # Run every algorithm and update the temporary results
            start_time = time.time()
            d = qp.find_quickest_path(qp.dijkstra_algorithm, source, target)
            temp_results['D']['runtimes'].append(time.time()-start_time)
            temp_results['D']['explored_nodes'].append(d[0]['explored_nodes'])
            start_time = time.time()
            dl = qp.find_quickest_path(qp.dijkstra_landmarks_algorithm, source,
                                       target)
            temp_results['DL']['runtimes'].append(time.time()-start_time)
            temp_results['DL']['explored_nodes'].append(dl[0]['explored_nodes'])
            start_time = time.time()
            alt1 = qp.find_quickest_path(qp.astar_algorithm, source, target,
                                         nw.landmark_heuristic_1)
            temp_results['ALT1']['runtimes'].append(time.time()-start_time)
            temp_results['ALT1']['explored_nodes'].append(alt1[0]['explored_nodes'])
            start_time = time.time()
            alt2 = qp.find_quickest_path(qp.astar_algorithm, source, target,
                                         nw.landmark_heuristic_2)
            temp_results['ALT2']['runtimes'].append(time.time()-start_time)
            temp_results['ALT2']['explored_nodes'].append(alt2[0]['explored_nodes'])
            start_time = time.time()
            bdd = qp.find_quickest_path(qp.bidirectional_dijkstra_algorithm,
                                        source, target)
            temp_results['BDD']['runtimes'].append(time.time()-start_time)
            temp_results['BDD']['explored_nodes'].append(bdd[0]['explored_nodes'])
            start_time = time.time()
            bdalt1 = qp.find_quickest_path(qp.bidirectional_astar_algorithm,
                                                    source, target,
                                                    nw.landmark_heuristic_1)
            temp_results['BDALT1']['runtimes'].append(time.time()-start_time)
            temp_results['BDALT1']['explored_nodes'].append(bdalt1[0]['explored_nodes'])
            start_time = time.time()
            bdalt2 = qp.find_quickest_path(qp.bidirectional_astar_algorithm,
                                                    source, target,
                                                    nw.landmark_heuristic_2)
            temp_results['BDALT2']['runtimes'].append(time.time()-start_time)
            temp_results['BDALT2']['explored_nodes'].append(bdalt2[0]['explored_nodes'])
            if EUCLIDEAN_DATASET == graph_type:
                start_time = time.time()
                aed = qp.find_quickest_path(qp.astar_algorithm, source, target,
                            nw.euclidean_distance)
                temp_results['AED']['runtimes'].append(time.time()-start_time)
                temp_results['AED']['explored_nodes'].append(aed[0]['explored_nodes'])
                start_time = time.time()
                bdaed = qp.find_quickest_path(qp.bidirectional_astar_algorithm,
                                                        source, target,
                                                        nw.euclidean_distance)
                temp_results['BDAED']['runtimes'].append(time.time()-start_time)
                temp_results['BDAED']['explored_nodes'].append(bdaed[0]['explored_nodes'])

        # Update the results
        results['D']['runtimes'].append(nw.statistics.pmean(temp_results['D']['runtimes']))
        results['D']['explored_nodes'].append(nw.statistics.pmean(temp_results['D']['explored_nodes']))
        results['D']['explored_space'].append(results['D']['explored_nodes'][-1]/len(graph.nodes()))
        results['DL']['runtimes'].append(nw.statistics.pmean(temp_results['DL']['runtimes']))
        results['DL']['explored_nodes'].append(nw.statistics.pmean(temp_results['DL']['explored_nodes']))
        results['DL']['explored_space'].append(results['DL']['explored_nodes'][-1]/len(graph.nodes()))
        results['ALT1']['runtimes'].append(nw.statistics.pmean(temp_results['ALT1']['runtimes']))
        results['ALT1']['explored_nodes'].append(nw.statistics.pmean(temp_results['ALT1']['explored_nodes']))
        results['ALT1']['explored_space'].append(results['ALT1']['explored_nodes'][-1]/len(graph.nodes()))
        results['ALT2']['runtimes'].append(nw.statistics.pmean(temp_results['ALT2']['runtimes']))
        results['ALT2']['explored_nodes'].append(nw.statistics.pmean(temp_results['ALT2']['explored_nodes']))
        results['ALT2']['explored_space'].append(results['ALT2']['explored_nodes'][-1]/len(graph.nodes()))
        results['BDD']['runtimes'].append(nw.statistics.pmean(temp_results['BDD']['runtimes']))
        results['BDD']['explored_nodes'].append(nw.statistics.pmean(temp_results['BDD']['explored_nodes']))
        results['BDD']['explored_space'].append(results['BDD']['explored_nodes'][-1]/len(graph.nodes()))
        results['BDALT1']['runtimes'].append(nw.statistics.pmean(temp_results['BDALT1']['runtimes']))
        results['BDALT1']['explored_nodes'].append(nw.statistics.pmean(temp_results['BDALT1']['explored_nodes']))
        results['BDALT1']['explored_space'].append(results['BDALT1']['explored_nodes'][-1]/len(graph.nodes()))
        results['BDALT2']['runtimes'].append(nw.statistics.pmean(temp_results['BDALT2']['runtimes']))
        results['BDALT2']['explored_nodes'].append(nw.statistics.pmean(temp_results['BDALT2']['explored_nodes']))
        results['BDALT2']['explored_space'].append(results['BDALT2']['explored_nodes'][-1]/len(graph.nodes()))
        if EUCLIDEAN_DATASET == graph_type:
            results['AED']['runtimes'].append(nw.statistics.pmean(temp_results['AED']['runtimes']))
            results['AED']['explored_nodes'].append(nw.statistics.pmean(temp_results['AED']['explored_nodes']))
            results['AED']['explored_space'].append(results['AED']['explored_nodes'][-1]/len(graph.nodes()))
            results['BDAED']['runtimes'].append(nw.statistics.pmean(temp_results['BDAED']['runtimes']))
            results['BDAED']['explored_nodes'].append(nw.statistics.pmean(temp_results['BDAED']['explored_nodes']))
            results['BDAED']['explored_space'].append(results['BDAED']['explored_nodes'][-1]/len(graph.nodes()))

    return results


def store_data(data, filename):
    '''Pickles the argument data as filename.pckl.'''
    filename = '%s.pckl' % filename
    f = open(filename, 'w')
    pickle.dump(data, f)
    f.close()


def load_data(filename):
    '''Loads the pickled file with the argument filename.pckl.'''
    filename = '%s.pckl' % filename
    f = open(filename)
    data = pickle.load(f)
    f.close()
    return data


def main():
    # Landmark Selection
    graph = nw.get_noneuclidean_dataset(floors=5)
    graph.set_random_travel_times()  # Set edge weights
    landmark_selection(graph, 20, 500)

    # Optimality
    graph = nw.get_euclidean_dataset(n=1015, radius=1.1)
    print len(graph.nodes()), len(graph.edges())
    landmarks = nw.Landmarks(graph)
    landmarks.build_landmarks(15, nw.centrality_selection)
    optimality(graph, landmarks, 1000, 1)
    print

    #print 'Reading'
    # The cumulative size of the four datasets is over 2GB
    # * euclidean dataset (1-20) and (1-40)
    # * non-euclidean dataset (1-20) and (1-40)
    #datasets = load_data('euclidean_datasets_40')
    #print 'Loaded into memory'
    #print

    # Runtime, explored vertices, and explored space
    run_simulation(dataset_size=20, graph_type=EUCLIDEAN_DATASET,
                   n_landmarks=15, selection_algorithm=nw.centrality_selection,
                   iterator_type=WHEELCHAIRS, max_searches=5,
                   max_num_wheelchairs=2000, increment=100)
    run_simulation(dataset_size=40, graph_type=EUCLIDEAN_DATASET,
                   n_landmarks=15, selection_algorithm=nw.centrality_selection,
                   iterator_type=GRAPH_SIZE, max_searches=5,
                   max_num_wheelchairs=100)


if __name__ == '__main__':
    main()

    # Safe exit
    sys.exit(0)
