# -*- coding: utf-8 -*-

'''astar.py: Contains an implementation of the A* algorithm and some
heuristics.

The A* implementation is a modifed version of the NetworkX -- a network/graph
library for Python -- implementation.'''

import heapq
import math


def landmark_heuristic_1(u, v, landmarks):
    '''Returns an approximate of the lower bound distance between u and v.
    It does this by utilizing the triangle inequality of u, v, and the closest
    landmark (l) to u: max{d(l, v) - d(u, l), 0}.'''
    n_lndmrk = landmarks.nearest_landmark_query(u)
    dist, _ = landmarks.shortest_path_query(n_lndmrk)
    return max(dist[v]-dist[u], 0)


def landmark_heuristic_2(u, v, landmarks):
    '''Returns an approximate of the lower bound distance between u and v that
    is tighter but more computationally intensive than landmark_heuristic_1().
    It applies the ideas from landmark_heuristic_1() symmetrically by also
    locating the nearest landmark of v.'''
    n_lndmrk_u = landmarks.nearest_landmark_query(u)
    n_lndmrk_v = landmarks.nearest_landmark_query(v)
    u_dist, _ = landmarks.shortest_path_query(n_lndmrk_u)
    v_dist, _ = landmarks.shortest_path_query(n_lndmrk_v)
    return max(u_dist[v]-u_dist[u], v_dist[u]-v_dist[v], 0)


def euclidean_distance(u, v, landmarks=None):
    '''Returns an approximate of the lower bound distance between u and v by
    calculating the Euclidean distance. The distance is multiplied by a very
    low number in an effort to break ties between paths with the same f
    value.
    Please note that landmarks is not used but is needed to generalize for
    the two other heuristics.'''
    dx = abs(u.pos[0]-v.pos[0])
    dy = abs(u.pos[1]-v.pos[1])
    return math.sqrt(dx*dx+dy*dy) * (1.0 + 0.00001)


def astar(graph, source, target, heuristic, landmarks=None):
    '''Returns the shortest path and the length of that path from source to
    target. The path is calculated using the well known informed best-first
    search algorithm called A*. The algorithm must be supplied with a
    heuristic function which approximates the distance between the current
    node and the target node. Landmarks can be supplied and used in conjunction
    with the heuristic function.'''
    # Base case when source == target
    if source == target:
        return {source: (0, 0)}, [source], 0

    explored_set = {}
    enqueued = {}
    # Fringe is defined as: (f, hash(current_node), g, current_node, parent)
    fringe = []  # The currently seen but otherwise unexplored nodes
    heapq.heappush(fringe, (0, hash(source), 0, source, None))

    while fringe:
        # Pop the node, n, with the lowest f(n) value, where f(n) = g(n) + h(n)
        _, __, dist, current_node, parent = heapq.heappop(fringe)

        # Exit if we have found the target
        if current_node == target:
            explored_set[current_node] = parent
            break

        # Skip already explored nodes as we already knows its neighbours
        if current_node in explored_set:
            continue

        # Mark the explored node
        explored_set[current_node] = parent

        # Retrieve and iterate through the neighbours of the current node
        edges = iter(graph[current_node].items())
        for neighbour, data in edges:
            if neighbour in explored_set:
                continue
            new_dist = dist + data.get('weight', 1)  # Let 1 be default value
            if neighbour in enqueued:
                q_dist, h = enqueued[neighbour]
                # Assumes consistent heuristic and just continue the search
                # if we visit the node (neighbour) a second time
                if q_dist <= new_dist:
                    continue
            else:
                h = heuristic(neighbour, target, landmarks)

            enqueued[neighbour] = (new_dist, h)
            heapq.heappush(fringe, (new_dist + h, hash(neighbour), new_dist,
                                    neighbour, current_node))

    # Remember that current_node may or may not be the target node
    # Return (dist, path, explored_nodes)
    return enqueued, reconstruct_path(current_node, parent, explored_set), len(explored_set)


def reconstruct_path(current_node, parent, explored_set):
    '''Reconstructs the path from current node to root (or any node without
    a parent) by iteratively traversing the explored set (via parents).'''
    path = [current_node]
    node = parent
    while node:
        path.append(node)
        node = explored_set[node]

    path.reverse()
    return path
