# -*- coding: utf-8 -*-

'''bidirectional.py: Contains an implementation of the bi-directional
Dijkstra's and A* algorithm.

This implementation of bi-directional A* search is inspired by the
bi-directional dijkstra algorithm in NetworkX -- a network/graph library for
Python.'''

import heapq

import networkx as nx

import astar


def bidirectional_astar(graph, source, target, heuristic, landmarks=None):
    '''Returns the shortest path and the length of that path from source to
    target. The path is calculated using the bi-directional A* algorithm.
    The algorithm is implemented by running two A* searches
    quasi-simultaneously -- one from source to target and another from target
    to source. This is done by alternating between the forward and backward
    A* search. One heuristic is applied to both searches.'''
    # Base case when source == target
    if source == target:
        return 0, [source], 0

    direction = 1  # forward: 0, backward: 1
    explored_set = [{}, {}]  # [{forward}, {backward}]
    enqueued = [{}, {}]  # [{forward}, {backward}]
    # Fringe is defined as: (f, hash(current_node), g, current_node, parent)
    fringe_f, fringe_b = [], []
    heapq.heappush(fringe_f, (0, hash(source), 0, source, None))
    heapq.heappush(fringe_b, (0, hash(target), 0, target, None))

    while fringe_f and fringe_b:
        # Find out if we are doing forward or backward search this turn
        direction = 1-direction
        # Pop the node, n, with the lowest f(n) value, where f(n) = g(n) + h(n)
        if direction:
            _, __, dist, current_node, parent = heapq.heappop(fringe_b)
        else:
            _, __, dist, current_node, parent = heapq.heappop(fringe_f)

        # Skip already explored nodes as we already knows its neighbours
        if current_node in explored_set[direction]:
            continue

        # Mark the explored node
        explored_set[direction][current_node] = parent

        # Exit if current_node has been explored by forward and backward search
        if current_node in explored_set[1-direction]:
            break

        # Retrieve and iterate through the neighbours of the current node
        edges = iter(graph[current_node].items())
        for neighbour, data in edges:
            if neighbour in explored_set[direction]:
                continue
            new_dist = dist + data.get('weight', 1)  # Let 1 be default value
            if neighbour in enqueued[direction]:
                q_dist, h = enqueued[direction][neighbour]
                # Assumes consistent heuristic and just continue the search
                # if we visit the node (neighbour) a second time
                if q_dist <= new_dist:
                    continue
            else:
                if direction:
                    h = heuristic(neighbour, source, landmarks)
                else:
                    h = heuristic(neighbour, target, landmarks)

            enqueued[direction][neighbour] = (new_dist, h)
            if direction:
                heapq.heappush(fringe_b, (new_dist + h, hash(neighbour),
                               new_dist, neighbour, current_node))
            else:
                heapq.heappush(fringe_f, (new_dist + h, hash(neighbour),
                               new_dist, neighbour, current_node))

    # Concatenate and return the results from the forward and backward search
    path_f = astar.reconstruct_path(current_node,
                                    explored_set[0][current_node],
                                    explored_set[0])
    path_b = astar.reconstruct_path(current_node,
                                    explored_set[1][current_node],
                                    explored_set[1])[::-1][1:]

    # Return (dist, path, explored_nodes)
    if current_node in enqueued[0] and current_node in enqueued[1]:
        return enqueued[0][current_node][0]+enqueued[1][current_node][0],\
               path_f+path_b, len(explored_set[0])+len(explored_set[1])
    elif current_node in enqueued[0]:
        return enqueued[0][current_node][0], path_f+path_b,\
               len(explored_set[0])+len(explored_set[1])
    elif current_node in enqueued[1]:
        return enqueued[1][current_node][0], path_f+path_b,\
               len(explored_set[0])+len(explored_set[1])
    else:
        return 0, [], 0


def bidirectional_dijkstra(graph, source, target, weight='weight'):
    '''Returns the shortest path and the length of that path from source to
    target. The path is calculated using the bi-directional dijkstra algorithm
    from the NetworkX package. This version is more stable than bi-directional
    A* with zero heuristic (due to less bookkeeping). The algorithm is
    provided in full here (instead of a function call) just to be able to
    return the number of explored nodes.'''
    if source == target: return (0, [source], 0)
    #Init:   Forward             Backward
    dists =  [{},                {}]# dictionary of final distances
    paths =  [{source:[source]}, {target:[target]}] # dictionary of paths
    fringe = [[],                []] #heap of (distance, node) tuples for extracting next node to expand
    seen =   [{source:0},        {target:0} ]#dictionary of distances to nodes seen
    #initialize fringe heap
    heapq.heappush(fringe[0], (0, source))
    heapq.heappush(fringe[1], (0, target))
    #neighs for extracting correct neighbor information
    if graph.is_directed():
        neighs = [graph.successors_iter, graph.predecessors_iter]
    else:
        neighs = [graph.neighbors_iter, graph.neighbors_iter]
    #variables to hold shortest discovered path
    #finaldist = 1e30000
    finalpath = []
    dir = 1
    while fringe[0] and fringe[1]:
        # choose direction
        # dir == 0 is forward direction and dir == 1 is back
        dir = 1-dir
        # extract closest to expand
        (dist, v )= heapq.heappop(fringe[dir])
        if v in dists[dir]:
            # Shortest path to v has already been found
            continue
        # update distance
        dists[dir][v] = dist #equal to seen[dir][v]
        if v in dists[1-dir]:
            # if we have scanned v in both directions we are done
            # we have now discovered the shortest path
            return (finaldist,finalpath, len(dists[0])+len(dists[1]))

        for w in neighs[dir](v):
            if(dir==0): #forward
                if graph.is_multigraph():
                    minweight=min((dd.get(weight,1)
                               for k,dd in graph[v][w].items()))
                else:
                    minweight=graph[v][w].get(weight,1)
                vwLength = dists[dir][v] + minweight #graph[v][w].get(weight,1)
            else: #back, must remember to change v,w->w,v
                if graph.is_multigraph():
                    minweight=min((dd.get(weight,1)
                               for k,dd in graph[w][v].items()))
                else:
                    minweight=graph[w][v].get(weight,1)
                vwLength = dists[dir][v] + minweight #graph[w][v].get(weight,1)

            if w in dists[dir]:
                if vwLength < dists[dir][w]:
                    raise ValueError("Contradictory paths found: negative weights?")
            elif w not in seen[dir] or vwLength < seen[dir][w]:
                # relaxing
                seen[dir][w] = vwLength
                heapq.heappush(fringe[dir], (vwLength,w))
                paths[dir][w] = paths[dir][v]+[w]
                if w in seen[0] and w in seen[1]:
                    #see if this path is better than than the already
                    #discovered shortest path
                    totaldist = seen[0][w] + seen[1][w]
                    if finalpath == [] or finaldist > totaldist:
                        finaldist = totaldist
                        revpath = paths[1][w][:]
                        revpath.reverse()
                        finalpath = paths[0][w] + revpath[1:]
    raise nx.NetworkXNoPath("No path between %s and %s." % (source, target))
