# -*- coding: utf-8 -*-

'''dijkstra.py: Contains an implementation of the Dijkstra's algorithm.

This implementation of Dijkstra's algorithm is a modifed version of the
NetworkX -- a network/graph library for Python -- implementation.'''

import heapq


def dijkstra(graph, source, target=None, landmarks=None):
    '''Solves the single-source shortest path problem; however, if the target
    is provided the single-pair shortest path problem is solved instead. The
    path is calculated using the well known graph search algorithm Dijkstra's
    algorithm. This implementation can use landmarks as shortcuts, however
    this might not yield optimal paths (upper bound).'''
    # Base case when source == target
    if target and source == target:
        return {source: 0}, {source: [source]}, 0

    # Setup landmarks data structures if necessary
    if landmarks:
        best_landmark = {}

    dist = {}  # distances from source to the key node
    paths = {source: [source]}  # paths from source to the key node
    seen = {source: 0}
    # Fringe is defined as (dist, current_node)
    fringe = []  # The currently seen but otherwise unexplored nodes
    heapq.heappush(fringe, (0, source))

    while fringe:
        # Pop the node, n, with the lowest distance
        current_dist, current_node = heapq.heappop(fringe)

        # Skip already explored nodes as we already knows its neighbours
        if current_node in dist:
            continue

        # Set the current distance for the current node
        dist[current_node] = current_dist

        # Exit if we have found the target (single-pair version only)
        if target and current_node == target:
            break

        # Landmarks can be used in the single-pair version as a means to
        # jump quickly around the search space (used as shortcuts)
        if target and landmarks and current_node in landmarks.landmark_nodes:
            # Exit if distance to the currently best landmark is lower than the
            # distance to this landmark (better upper bound)
            if best_landmark and best_landmark['d'] < dist[current_node]:
                return best_landmark['dist'], best_landmark['paths'], len(best_landmark)
            else:
                best_landmark = update_best_landmark(current_node, target,
                                                     dist, paths, landmarks)

        # Retrieve and iterate through the neighbours of the current node
        edges = iter(graph[current_node].items())
        for neighbour, data in edges:
            new_dist = dist[current_node] + data.get('weight', 1)  # Let 1 be default value
            if neighbour in dist and new_dist < dist[neighbour]:
                raise ValueError('Contradiction has occured')
            elif neighbour not in seen or new_dist < seen[neighbour]:
                seen[neighbour] = new_dist
                heapq.heappush(fringe, (new_dist, neighbour))
                paths[neighbour] = paths[current_node] + [neighbour]

    # Return (dist, paths, explored_nodes)
    return dist, paths, len(dist)


def update_best_landmark(current_node, target, dist, paths, landmarks):
    '''Utility function that returns information about the landmark:
    current_node.'''
    # Find the landmark distance/path to the target
    l_dist, l_paths = landmarks.shortest_path_query(current_node)

    # Concatenate the correct distance values
    _dist = dict(dist)
    tmp_dist = _dist[current_node]
    for node in l_paths[target]:
        _dist[node] = tmp_dist + l_dist[node]

    # Concatenate the correct nodes
    _paths = dict(paths)
    _paths[target] = _paths[current_node] +\
                     [node for node in l_paths[target] if node != current_node]

    # Return the currently best landmark -> target
    return {'d': dist[current_node], 'current_node': current_node,
            'dist': _dist, 'paths': _paths}
