# -*- coding: utf-8 -*-

'''landmarks.py: Contains an implementation of the landmark data structure and
some selection functions.'''

import random

import networkx as nx

import dijkstra
import statistics


def uniform_selection(graph, n):
    '''Returns n uniformly selected nodes from the graph.'''
    nodes = set()
    while n <= len(graph.nodes()) and len(nodes) != n:
        nodes.add(random.choice(graph.nodes()))
    return list(nodes)


def degree_selection(graph, n):
    '''Returns n nodes with the highest vertex degree from the graph.'''
    return statistics.n_largest_vertex_degree(graph, n)[1]


def centrality_selection(graph, n):
    '''Returns n nodes with the highest eigenvector centrality from the graph.
    '''
    return statistics.n_largest_centrality(graph,
                                           nx.eigenvector_centrality_numpy,
                                           n)[1]


def coverage_selection(graph, n):
    '''Returns n nodes with the highest vertex coverage. A node n is said to be
    covering a vertex pair (u, v) if n lies on the shortest path between u and
    v.'''
    coverage = dict.fromkeys(graph.nodes(), 0)

    # Sample the shortest path of half of the nodes in the graph
    #nodes = uniform_selection(graph, len(graph.nodes())/2)
    nodes = uniform_selection(graph, n*2)
    for node in nodes:
        dist, paths, _ = dijkstra.dijkstra(graph, node)
        for cover in coverage:
            coverage[cover] += len([x[1:] for x in paths.values()
                                    if cover in x[1:]])

    return statistics.n_largest(coverage, n)[1]


class Landmarks(object):
    '''Data structure for the acceleration method called landmarks. A landmark
    contains shortest path information from it to every other node in the graph
    (single-source).'''
    def __init__(self, graph):
        self.graph = graph
        self.landmarks = {}
        self.landmark_nodes = []
        self.nearest_landmark = {}

    def build_landmarks(self, n, selection_algorithm):
        '''Builds n landmarks using the provided selection algorithm.'''
        # Check if there are enough nodes in the graph to create landmarks
        if len(self.graph.nodes()) < n:
            raise ValueError('Amount of landmarks exceeds the amount of nodes\
                             available')

        # Clear existing landmarks if necessay
        if self.landmarks:
            print 'Landmarks already exist, clearing...'
            self.landmarks = {}

        # Find landmark nodes
        self.landmark_nodes = selection_algorithm(self.graph, n)

        # Compute and store single-source shortest path data for each landmark
        for node in self.landmark_nodes:
            dist, paths, _ = dijkstra.dijkstra(self.graph, node)
            self.landmarks[node] = {'dist': dist, 'paths': paths}

        # Pre-compute the nearest landmark for all nodes in the graph
        for node in self.graph.nodes():
            current_nearest = None
            for landmark in self.landmark_nodes:
                dist, paths = self.shortest_path_query(self.landmark_nodes[0])
                if current_nearest:
                    if dist[node] < current_nearest[1][node]:
                        current_nearest = landmark, dist
                else:
                    current_nearest = landmark, dist
            self.nearest_landmark[node] = current_nearest[0]


    def nearest_landmark_query(self, node):
        '''Returns the nearest landmark to the argument node as
        (landmark, dist).'''
        return self.nearest_landmark[node]

    def shortest_path_query(self, source):
        '''Returns the single-source shortest path, where source is a
        landmark.'''
        # Check if source is a landmark
        if not source in self.landmark_nodes:
            raise ValueError('Source node is not a landmark')

        # Query and return shortest path and distance data
        return self.landmarks[source]['dist'],\
               self.landmarks[source]['paths']
