# -*- coding: utf-8 -*-

'''quickest_path.py: Contains an implementation of the quickest path between
a source and target node via one wheelchair of a set of wheelchairs.'''

import time

import astar
import bidirectional
import dijkstra


class Quickest_path(object):
    '''Object for computing the quickest path between a source and target
    node via one wheelcahir from self.wheelchairs.'''
    def __init__(self, graph, wheelchairs, landmarks=None):
        self.graph = graph
        self.wheelchairs = wheelchairs
        self.landmarks = landmarks

    def find_quickest_path(self, algorithm, source, target, heuristic=None):
        '''Returns the quickest path between a source and target node via
        one wheelchair from self.wheelchairs. The result is a ranked list
        of the nearest wheelchair with respect to the path between source
        and target. Each element contains the node where the wheelchair
        resides, the distance and path from source -> wheelchair -> target,
        the number of explored nodes, and the runtime.
        The function needs the name of the algorithm (and any heuristic) to
        be able to run.'''
        # search_result is defined as:
        # {wheelchair: {dist: ?, path: ?, explored_nodes: ?}}
        search_result = []
        for wheelchair in self.wheelchairs:
            start_time = time.time()
            # Compute source -> wheelchair
            if heuristic:
                _result1 = algorithm(source, wheelchair, heuristic)
            else:
                _result1 = algorithm(source, wheelchair)

            # Compute wheelchair -> target
            if heuristic:
                _result2 = algorithm(wheelchair, target, heuristic)
            else:
                _result2 = algorithm(wheelchair, target)

            # Update search result
            search_result.append({'wheelchair': wheelchair,
                                  'dist': _result1[0] + _result2[0],
                                  'path': _result1[1] + _result2[1][1:],
                                  'explored_nodes': _result1[2] + _result2[2],
                                  'runtime': time.time()-start_time})

        # Sort and return the search results
        return sorted(search_result, key=lambda k: k['dist'])

    def dijkstra_algorithm(self, source, target):
        d, p, e = dijkstra.dijkstra(self.graph, source, target)
        return (d[target], p[target], e)

    def dijkstra_landmarks_algorithm(self, source, target):
        d, p, e = dijkstra.dijkstra(self.graph, source, target, self.landmarks)
        return (d[target], p[target], e)

    def astar_algorithm(self, source, target, heuristic):
        d, p, e = astar.astar(self.graph, source, target, heuristic, self.landmarks)
        return (d[target][0], p, e)

    def bidirectional_dijkstra_algorithm(self, source, target):
        d, p, e = bidirectional.bidirectional_dijkstra(self.graph, source, target)
        return (d, p, e)

    def bidirectional_astar_algorithm(self, source, target, heuristic):
        d, p, e = bidirectional.bidirectional_astar(self.graph, source, target,
                                                    heuristic, self.landmarks)
        return (d, p, e)
