# -*- coding: utf-8 -*-

'''statistics.py: Contains a few statistical measurement functions.'''

import heapq
import math


def pmean(population):
    '''Computes the population mean.'''
    return sum(population)/len(population)


def pvariance(population, mean=None):
    '''Computes the population variance. A larger variance indicates that the
    population data is spread out; a small variance indicates it is clustered
    closely around the mean.'''
    if not mean:
        mean = pmean(population)

    return sum([((p - mean)**2)/len(population) for p in population])


def pstdev(population, mean=None):
    '''Computes the squared population variance.'''
    return math.sqrt(pvariance(population, mean))


def variance(sample, mean=None):
    '''Computes the sample variance. A sample is defined as being part of a
    population, e.g. it might just be 1% of the whole population.'''
    if not mean:
        mean = pmean(sample)

    return sum([((d - mean)**2)/(len(sample)-1) for d in sample])


def stdev(sample, mean=None):
    '''Computes the squared sample variance.'''
    return math.sqrt(variance(sample, mean))


def graph_cardinality(graph):
    '''Returns the length (cardinality) of the vertex and edge set.'''
    return len(graph.nodes()), len(graph.edges())


def n_largest_vertex_degree(graph, n=None):
    '''Returns the n largest vertex degrees.'''
    return n_largest(graph.degree(graph.nodes()), n)


def n_smallest_vertex_degree(graph, n=None):
    '''Returns the n smallest vertex degrees.'''
    return n_smallest(graph.degree(graph.nodes()), n)


def n_largest_centrality(graph, centrality, n=None):
    '''Returns the n largest centralities.'''
    return n_largest(centrality(graph), n)


def n_largest(dictionary, n=None):
    '''Returns the n largest elements of the dictionary as
    (dictionary, [keys]). If n=None then the maximum element is returned as
    (value, [keys]) instead.'''
    if n:
        return dictionary, heapq.nlargest(n, dictionary, dictionary.get)
    else:
        nodes = [k for k, v in dictionary.iteritems() if v == max(dictionary.values())]
        return dictionary[max(dictionary, key=dictionary.get)], nodes


def n_smallest(dictionary, n=None):
    '''Returns the n smallest elements of the dictionary as
    (dictionary, [keys]). If n=None then the minimum element is returned as
    (value, [keys]) instead.'''
    if n:
        return dictionary, heapq.nsmallest(n, dictionary, dictionary.get)
    else:
        nodes = [k for k, v in dictionary.iteritems() if v == min(dictionary.values())]
        return dictionary[min(dictionary, key=dictionary.get)], nodes
