# -*- coding: utf-8 -*-

'''dataset_definition.py: Contains the graph definitions used by the datasets.

The graph definition allows for setting edge weights (travel time) and
placement of wheelchairs.'''

import random
import string

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np

TT_STAIRS = 3
TT_ELEVATORS = 2
TT_MEDIUM = 1
TT_INSTANT = 0


class Hospital_graph(nx.Graph):
    '''Simple extension to the networkx.Graph class.'''
    def select_random_node(self):
        '''Standardized way to select a random node from the graph.'''
        return random.choice(self.nodes())

    def place_n_wheelchairs(self, num_wheelchairs, n_types=1):
        '''Distributes num_wheelchairs randomly among the graph's nodes.
        n_types allows for different wheelchair types.'''
        i = 0
        ids = []
        wheelchairs = []
        while i < int(num_wheelchairs):
            n = self.select_random_node()
            while '.S' in n.label:  # Do not place wheelchairs in stairs
                n = self.select_random_node()
            wheelchairs.append(n)
            new_id = id_generator(10)
            while new_id in ids:
                new_id = id_generator(10)
            w_type = random.randint(1, n_types)
            n.wheelchairs.append(Wheelchair(new_id, w_type))
            i += 1

        return wheelchairs

    def set_random_travel_times(self):
        '''Randomizes all edge weights if endpoints are specified as either
        S (stair), E (elevator), C (corridor), or R (room). The edge weight
        symbolizes travel time.'''
        try:
            t = self.edges(data=True)[0][0].label
            if not ('S' in t or 'E' in t or 'C' in t or 'R' in t):
                return  False
        except Exception:
            # This does not matter in our simple case
            return False

        for edge in self.edges(data=True):
            edge[2]['weight'] = 0.0  # ensure weight existence
            if 'S' in edge[0].label and 'S' in edge[1].label:
                edge[2]['weight'] = travel_time_generator(TT_STAIRS)
            elif 'E' in edge[0].label and 'E' in edge[1].label:
                edge[2]['weight'] = travel_time_generator(TT_ELEVATORS)
            elif ('S' in edge[0].label != 'S' in edge[1].label) or\
                 ('E' in edge[0].label != 'E' in edge[1].label):
                edge[2]['weight'] = travel_time_generator(TT_INSTANT)
            else:
                edge[2]['weight'] = travel_time_generator(TT_MEDIUM)

    def reset_wheelchair_placement(self):
        '''Remove all wheelchairs from the graph.'''
        for room in self.nodes():
            room.wheelchairs = []

    def reset_edge_weights(self):
        '''Sets all edge weights to zero.'''
        for edge in self.edges(data=True):
            edge[2]['weight'] = 0.0

    def pretty_print_edges(self):
        '''Print all edges, one for each line.'''
        for a, b, weight in self.edges(data=True):
            print '%s <-> %s: %s' % (a, b, weight['weight'])

    def draw(self, radius=None):
        '''Draw and display graph using matplotlib.pyplot.'''
        plt.figure(figsize=(8, 8), facecolor='white')
        if radius:
            pos = {}
            for node in self.nodes(data=True):
                pos[node[0]] = node[0].pos

            circle = plt.Circle((0, 0), radius=radius, alpha=0.05)
            axis = plt.gca()
            axis.add_patch(circle)
        else:
            pos = nx.spring_layout(self)  # default networkx layout

        p = nx.single_source_shortest_path_length(self, self.select_random_node())
        nx.draw_networkx_nodes(self, pos, nodelist=p.keys(), node_size=80,
                               node_color=p.values(), cmap=plt.cm.Blues)
        nx.draw_networkx_edges(self, pos, alpha=0.4)
        plt.axis('off')
        plt.show()

class Node(object):
    '''A node must be able to hold wheelchairs and should be recognized by a
    label.'''
    def __init__(self, label):
        self.label = str(label)
        self.pos = np.array([0.0, 0.0])
        self.wheelchairs = []

    def __str__(self):
        return self.label


class Wheelchair(object):
    '''A wheelchair must be uniquely defined by an id and can be specified as
    a specific type.'''
    def __init__(self, label=None, w_type=None):
        if not label:
            label = id_generator(10)
        self.label = label
        self.w_type = w_type

    def __str__(self):
        return self.label


def id_generator(size):
    '''Generates a random string id with argument size.'''
    chars = string.ascii_uppercase + string.ascii_lowercase + string.digits
    return ''.join(random.choice(chars) for x in xrange(size))


def travel_time_generator(tt_type=None):
    '''Generates a random travel time number from 1 to 120 seconds
    depending on the argument type.'''
    MIN, MAX = 1.0, 120.0
    if TT_STAIRS == tt_type:
        return random.uniform(30.0, MAX)  # 30 seconds to 2 minutes
    elif TT_ELEVATORS == tt_type:
        return random.uniform(30.0, 60.0)  # 30 to 60 seconds
    elif TT_MEDIUM == tt_type:
        return random.uniform(MIN, 60.0)  # 1 second to 1 minutes
    elif TT_INSTANT == tt_type:
        return MIN  # 1 second is regarded as instant travel time
    else:
        return random.uniform(MIN, MAX)
