# -*- coding: utf-8 -*-

'''datasets.py: Contains functions for retrieving two datasets.

Both datasets are modelled as graphs where the first resides in non-Euclidean
space and the other in Euclidean space.'''

import networkx as nx

import dataset_definition as datadef
import random_geometric_dataset as edataset
import ahl_dataset as nonedataset


def get_euclidean_dataset(filename=None, n=None, radius=None):
    '''Returns a randomly generated Euclidean dataset.'''
    # Dataset can be read from file
    if filename:
        # gpickle ensures that we have the exact same dataset
        return read_dataset(filename, 'gpickle')
    elif n and radius:
        return edataset.random_geometric_graph(n, radius, threshold=0.1,
                                               force_reachability=True)
    else:
        return datadef.Hospital_graph()


def get_noneuclidean_dataset(filename=None, floors=None):
    '''Returns a non-Euclidean dataset of the AHL (Akutten og
    Hjerte-lunge-senteret)building at St. Olavs Hospital.'''
    if filename:
        # gpickle ensures that we have the exact same dataset
        return read_dataset(filename, 'gpickle')
    elif floors:
        return nonedataset.ahl_graph(floors)
    else:
        return datadef.Hospital_graph()


def read_dataset(filename, filetype):
    '''Reads a graph dataset from disk.'''
    graph = datadef.Hospital_graph()
    name = '%s.%s' % (filename, filetype)
    try:
        if 'gpickle' == filetype:
            graph = nx.read_gpickle(name)
        elif 'gexf' == filetype:
            graph = nx.read_gexf(name, node_type=datadef.Node)
    except IOError, e:
        print 'IOError: %s' % e[1]

    return graph


def write_dataset(filename, graph, filetype):
    '''Write a graph dataset to disk.
    The .gexf file format should only be used as a human-readable file
    format.'''
    name = '%s.%s' % (filename, filetype)
    try:
        if 'gpickle' == filetype:
            nx.write_gpickle(graph, name)
        elif 'gexf' == filetype:
            nx.write_gexf(graph, name)
    except IOError, e:
        print 'IOError: %s' % e[1]
