# -*- coding: utf-8 -*-

'''random_geometric_dataset.py: Contains the randomly created geometric graph
datasets.'''

import networkx as nx
import numpy as np
import scipy.spatial as spatial

import dataset_definition as datadef


def random_geometric_graph(n, radius, threshold=0.1, force_reachability=False):
    '''Returns a geometric graph with n nodes.'''
    def build_graph(n, radius):
        '''Returns a random geometric graph with n nodes with two
        dimensions, (X, Y). The graph is built using k-d trees. An edge
        between u and v is created if the distance between them is less than
        the threshold value. radius represents the radius of the circle where
        positions are sampled from.

        The function is heavily inspired by this k-d tree implemention on
        stackoverflow: http://stackoverflow.com/questions/13792046/'''
        # Create the graph from the definition
        graph = datadef.Hospital_graph()

        # Sample n positions (X, Y) in a disc using rejection
        # (n*10.0 to ensure len(disc) == n)
        pos = np.random.uniform(low=-radius, high=radius, size=(n*10.0, 2.0))
        np.random.shuffle(pos)
        disc = np.array([p for p in pos if np.linalg.norm(p) < radius][0:n])

        # Add the nodes to the graph
        graph.add_nodes_from([datadef.Node(i) for i in range(n)])
        i = 0
        for node in graph.nodes():
            node.pos = disc[i]
            i += 1

        # Build k-d tree of points in the disc
        kd_tree = spatial.KDTree(disc)

        # Add edges to the graph by querying nodes from the k-d tree. There
        # exists an edge between two vertices if the distance between them
        # is less than the threshold value
        endpoint_pairs = kd_tree.query_pairs(threshold)
        graph.add_edges_from([(graph.nodes()[pair[0]], graph.nodes()[pair[1]])
                              for pair in endpoint_pairs])

        return graph

    # n must be positive
    if 1 > n:
        return datadef.Hospital_graph()

    # Build a geometric graph using k-d trees
    graph = build_graph(n, radius)

    # Brute-force approach to ensure node reachability
    if force_reachability:
        i = 0
        check = nx.single_source_shortest_path(graph, graph.nodes()[0]).keys()
        while len(check) != len(graph.nodes()) and i <= 100:
            graph = build_graph(n, radius)
            check = nx.single_source_shortest_path(graph, graph.nodes()[0]).keys()
            i += 1

        if len(check) != len(graph.nodes()):
            print 'Unable to ensure graph reachability'


    # Pre-compute Euclidean distances as edge weights
    for edge in graph.edges(data=True):
        edge[2]['weight'] = np.linalg.norm(edge[0].pos-edge[1].pos)

    return graph
